@@ -1461,13 +1461,15 @@ static void tone_map(pl_shader sh,
             .data = PL_TRANSPOSE_3X3(rgb2xyz.m),
         }));
 
-        // Tuned to meet the desired desaturation at 1000 -> SDR
-        float desat = dst_max > src_max ? 1.075f : 1.1f;
-        float exponent = logf(desat) / logf(1000 / PL_COLOR_SDR_WHITE);
-        GLSL("float orig = max(xyz.y, %s);                      \n"
-             "xyz.y = tone_map(xyz.y);                          \n"
-             "xyz.xz *= pow(xyz.y / orig, %s) * xyz.y / orig;   \n",
-             SH_FLOAT(dst_min), SH_FLOAT(exponent));
+        // Tuned to meet the desired desaturation at 1000 <-> SDR
+        float ratio = logf(src_max / dst_max) / logf(1000 / PL_COLOR_SDR_WHITE);
+        float coeff = src_max > dst_max ? 1 / 1.1f : 1.075f;
+        float desat = (coeff - 1) * fabsf(ratio) + 1;
+
+        GLSL("float orig = max(xyz.y, %s);  \n"
+             "xyz.y = tone_map(xyz.y);      \n"
+             "xyz.xz *= %s * xyz.y / orig;  \n",
+             SH_FLOAT(dst_min), SH_FLOAT(desat));
 
         // Extra luminance correction when reducing dynamic range
         if (src_max > dst_max)
@@ -1574,9 +1576,13 @@ static void adapt_colors(pl_shader sh,
         break;
 
     case PL_GAMUT_WARN:
-        GLSL("if (any(lessThan(color.rgb, vec3(-0.005))) ||     \n"
-             "    any(greaterThan(color.rgb, vec3(1.005))))     \n"
-             "    color.rgb = vec3(1.0, 0.0, 1.0); // magenta   \n");
+        GLSL("if (any(lessThan(color.rgb, vec3(-1e-6))) ||          \n"
+             "    any(greaterThan(color.rgb, vec3(1.0 + 1e-6))))    \n"
+             "{                                                     \n"
+             "    float k = dot(color.rgb, vec3(2.0 / 3.0));        \n"
+             "    color.rgb = clamp(vec3(k) - color.rgb, 0.0, 1.0); \n"
+             "    color.rgb = sqrt(color.rgb);                      \n"
+             "}                                                     \n");
         break;
 
     case PL_GAMUT_DARKEN: {
