@@ -666,6 +666,39 @@ static void pl_shader_tests(pl_gpu gpu)
         }));
     }
 
+    // Test deinterlacing
+    sh = pl_dispatch_begin(dp);
+    pl_shader_deinterlace(sh, pl_deinterlace_source( .cur = pl_field_pair(src) ), NULL);
+    REQUIRE(pl_dispatch_finish(dp, pl_dispatch_params(
+        .shader = &sh,
+        .target = fbo,
+    )));
+
+    // Test error diffusion
+    if (fbo->params.storable) {
+        for (int i = 0; i < pl_num_error_diffusion_kernels; i++) {
+            const struct pl_error_diffusion_kernel *k = pl_error_diffusion_kernels[i];
+            printf("testing error diffusion kernel '%s'\n", k->name);
+            sh = pl_dispatch_begin(dp);
+            bool ok = pl_shader_error_diffusion(sh, pl_error_diffusion_params(
+                .input_tex  = src,
+                .output_tex = fbo,
+                .new_depth  = 8,
+                .kernel     = k,
+            ));
+
+            if (!ok) {
+                fprintf(stderr, "kernel '%s' exceeds GPU limits, skipping...\n", k->name);
+                continue;
+            }
+
+            REQUIRE(pl_dispatch_compute(dp, pl_dispatch_compute_params(
+                .shader = &sh,
+                .dispatch_size = {1, 1, 1},
+            )));
+        }
+    }
+
     pl_dispatch_destroy(&dp);
     pl_tex_destroy(gpu, &src);
     pl_tex_destroy(gpu, &fbo);
@@ -978,6 +1011,11 @@ static void pl_render_tests(pl_gpu gpu)
     TEST_PARAMS(dither, temporal, true);
     TEST(cone_params, pl_cone_params, pl_vision_deuteranomaly, strength, 0);
 
+    // Test gamma-correct dithering
+    target.repr.bits.color_depth = 2;
+    TEST_PARAMS(dither, transfer, PL_COLOR_TRC_GAMMA22);
+    target.repr.bits.color_depth = 32;
+
     // Test HDR tone mapping
     image.color = pl_color_space_hdr10;
     TEST_PARAMS(color_map, tone_mapping_mode, PL_TONE_MAP_MODE_COUNT - 1);
@@ -1127,7 +1165,6 @@ static void pl_render_tests(pl_gpu gpu)
     struct pl_queue_params qparams = {
         .radius = pl_frame_mix_radius(&mix_params),
         .vsync_duration = 1.0 / 60.0,
-        .frame_duration = 1.0 / 24.0,
     };
 
     // Test large PTS jumps in frame mix
@@ -1142,11 +1179,13 @@ static void pl_render_tests(pl_gpu gpu)
 
     // Test mixer queue
 #define NUM_MIX_FRAMES 20
+    const float frame_duration = 1.0 / 24.0;
     struct pl_source_frame srcframes[NUM_MIX_FRAMES+1];
     srcframes[NUM_MIX_FRAMES] = (struct pl_source_frame) {0};
     for (int i = 0; i < NUM_MIX_FRAMES; i++) {
         srcframes[i] = (struct pl_source_frame) {
-            .pts = i * qparams.frame_duration,
+            .pts = i * frame_duration,
+            .duration = frame_duration,
             .map = frame_passthrough,
             .frame_data = &image,
         };
@@ -1159,7 +1198,7 @@ static void pl_render_tests(pl_gpu gpu)
     for (int i = 0; i < NUM_MIX_FRAMES; i++) {
         const struct pl_source_frame *src = &srcframes[i];
         if (i > 10) // test pushing in reverse order
-            src = &srcframes[NUM_MIX_FRAMES - 1 - i];
+            src = &srcframes[NUM_MIX_FRAMES + 10 - i];
         if (!pl_queue_push_block(queue, 1, src)) // mini-sleep
             pl_queue_push(queue, src); // push it anyway, for testing
     }
@@ -1185,7 +1224,6 @@ static void pl_render_tests(pl_gpu gpu)
     qparams = (struct pl_queue_params) {
         .radius = pl_frame_mix_radius(&mix_params),
         .vsync_duration = qparams.vsync_duration,
-        .frame_duration = qparams.frame_duration,
         .get_frame = get_frame_ptr,
         .priv = &frame_ptr,
     };
@@ -1202,6 +1240,28 @@ static void pl_render_tests(pl_gpu gpu)
     pl_queue_reset(queue);
     REQUIRE(pl_queue_update(queue, &mix, &qparams) == PL_QUEUE_EOF);
 
+    if (gpu->glsl.version >= 130) {
+        // Test deinterlacing
+        pl_queue_reset(queue);
+        printf("testing deinterlacing \n");
+        for (int i = 0; i < NUM_MIX_FRAMES; i++) {
+            struct pl_source_frame *src = &srcframes[i];
+            if (i > 10)
+                src = &srcframes[NUM_MIX_FRAMES + 10 - i];
+            src->first_field = PL_FIELD_EVEN;
+            pl_queue_push(queue, src);
+        }
+        pl_queue_push(queue, NULL);
+
+        qparams.pts = 0;
+        qparams.get_frame = NULL;
+        while ((ret = pl_queue_update(queue, &mix, &qparams)) != PL_QUEUE_EOF) {
+            REQUIRE(ret == PL_QUEUE_OK);
+            REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
+            qparams.pts += qparams.vsync_duration;
+        }
+    }
+
     pl_queue_destroy(&queue);
 
 error:
