@@ -1,15 +1,5 @@
 #include "gpu_tests.h"
-
-#ifndef EPOXY_HAS_EGL
-int main()
-{
-    return SKIP;
-}
-#else // EPOXY_HAS_EGL
-
 #include "opengl/utils.h"
-#include <epoxy/gl.h>
-#include <epoxy/egl.h>
 
 static void opengl_interop_tests(pl_gpu gpu)
 {
@@ -138,12 +128,16 @@ static void opengl_test_export_import(pl_opengl gl,
 
 int main()
 {
-    // Create the OpenGL context
-    if (!epoxy_has_egl_extension(EGL_NO_DISPLAY, "EGL_MESA_platform_surfaceless"))
+    if (!gladLoaderLoadEGL(EGL_NO_DISPLAY))
         return SKIP;
 
+    const char *extstr = eglQueryString(EGL_NO_DISPLAY, EGL_EXTENSIONS);
+    if (!extstr || !strstr(extstr, "EGL_MESA_platform_surfaceless"))
+        return SKIP;
+
+    // Create the OpenGL context
     EGLDisplay dpy = eglGetPlatformDisplayEXT(EGL_PLATFORM_SURFACELESS_MESA,
-                                              EGL_DEFAULT_DISPLAY, NULL);
+                                              (void *) EGL_DEFAULT_DISPLAY, NULL);
     if (dpy == EGL_NO_DISPLAY)
         return SKIP;
 
@@ -151,6 +145,9 @@ int main()
     if (!eglInitialize(dpy, &major, &minor))
         return SKIP;
 
+    if (!gladLoaderLoadEGL(dpy))
+        return SKIP;
+
     printf("Initialized EGL v%d.%d\n", major, minor);
     int egl_ver = major * 10 + minor;
 
@@ -238,6 +235,7 @@ int main()
             goto error;
 
         pl_opengl gl = pl_opengl_create(log, pl_opengl_params(
+            .get_proc_addr = (pl_voidfunc_t (*)(const char *)) eglGetProcAddress,
             .max_glsl_version = egl_vers[i].glsl_ver,
             .debug = true,
             .egl_display = dpy,
@@ -283,10 +281,9 @@ error: ;
     }
 
     eglTerminate(dpy);
+    gladLoaderUnloadEGL();
     pl_log_destroy(&log);
 
     if (!last_glsl.version)
         return SKIP;
 }
-
-#endif // EPOXY_HAS_EGL
