@@ -38,7 +38,6 @@ struct priv {
     struct pl_vulkan_swapchain_params params;
     VkSwapchainCreateInfoKHR protoInfo; // partially filled-in prototype
     VkSwapchainKHR swapchain;
-    VkSwapchainKHR old_swapchain;
     int cur_width, cur_height;
     int swapchain_depth;
     pl_rc_t frames_in_flight;       // number of frames currently queued
@@ -297,6 +296,9 @@ static void set_hdr_metadata(struct priv *p, const struct pl_hdr_metadata *metad
         .maxContentLightLevel = metadata->max_cll,
         .maxFrameAverageLightLevel = metadata->max_fall,
     });
+
+    // Keep track of applied HDR colorimetry metadata
+    p->color_space.hdr = *metadata;
 }
 
 pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
@@ -310,10 +312,9 @@ pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
         return NULL;
     }
 
-    struct pl_swapchain *sw = pl_zalloc_obj(NULL, sw, struct priv);
+    struct pl_swapchain_t *sw = pl_zalloc_obj(NULL, sw, struct priv);
     sw->impl = &vulkan_swapchain;
     sw->log = vk->log;
-    sw->ctx = sw->log;
     sw->gpu = gpu;
 
     struct priv *p = PL_PRIV(sw);
@@ -375,14 +376,8 @@ pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
                 vk_csp_name(p->formats.elem[i].colorSpace));
     }
 
-    struct pl_swapchain_colors hint = {0};
-    if (params->prefer_hdr) {
-        hint.primaries = PL_COLOR_PRIM_BT_2020;
-        hint.transfer = PL_COLOR_TRC_PQ;
-        hint.hdr = pl_hdr_metadata_hdr10;
-    }
-
     // Ensure there exists at least some valid renderable surface format
+    struct pl_color_space hint = {0};
     if (!pick_surf_format(sw, &hint))
         goto error;
 
@@ -402,6 +397,15 @@ static void vk_sw_destroy(pl_swapchain sw)
 
     pl_gpu_flush(gpu);
     vk_wait_idle(vk);
+
+    // Vulkan offers no way to know when a queue presentation command is done,
+    // leading to spec-mandated undefined behavior when destroying resources
+    // tied to the swapchain. Use an extra `vkQueueWaitIdle` on all of the
+    // queues we may have oustanding presentation calls on, to hopefully inform
+    // the driver that we want to wait until the device is truly idle.
+    for (int i = 0; i < vk->pool_graphics->num_queues; i++)
+        vk->QueueWaitIdle(vk->pool_graphics->queues[i]);
+
     for (int i = 0; i < p->images.num; i++)
         pl_tex_destroy(gpu, &p->images.elem[i]);
     for (int i = 0; i < p->sems.num; i++) {
@@ -548,11 +552,9 @@ error:
     return false;
 }
 
-static void destroy_swapchain(struct vk_ctx *vk, struct priv *p)
+static void destroy_swapchain(struct vk_ctx *vk, void *swapchain)
 {
-    assert(p->old_swapchain);
-    vk->DestroySwapchainKHR(vk->dev, p->old_swapchain, PL_VK_ALLOC);
-    p->old_swapchain = VK_NULL_HANDLE;
+    vk->DestroySwapchainKHR(vk->dev, vk_unwrap_handle(swapchain), PL_VK_ALLOC);
 }
 
 static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
@@ -564,35 +566,37 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
     VkImage *vkimages = NULL;
     uint32_t num_images = 0;
 
-    // It's invalid to trigger another swapchain recreation while there's more
-    // than one swapchain already active, so we need to flush any pending
-    // asynchronous swapchain release operations that may be ongoing
-    while (p->old_swapchain)
-        vk_poll_commands(vk, UINT64_MAX);
-
     VkSwapchainCreateInfoKHR sinfo = p->protoInfo;
-    sinfo.oldSwapchain = p->swapchain;
-
     if (!update_swapchain_info(p, &sinfo, w, h))
         return false;
 
-    PL_DEBUG(sw, "(Re)creating swapchain of size %dx%d",
-             sinfo.imageExtent.width,
-             sinfo.imageExtent.height);
-
-    VK(vk->CreateSwapchainKHR(vk->dev, &sinfo, PL_VK_ALLOC, &p->swapchain));
+#ifdef VK_EXT_full_screen_exclusive
+    // Explicitly disallow full screen exclusive mode if possible
+    static const VkSurfaceFullScreenExclusiveInfoEXT fsinfo = {
+        .sType = VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_INFO_EXT,
+        .fullScreenExclusive = VK_FULL_SCREEN_EXCLUSIVE_DISALLOWED_EXT,
+    };
+    if (vk->AcquireFullScreenExclusiveModeEXT)
+        vk_link_struct(&sinfo, &fsinfo);
+#endif
 
     p->suboptimal = false;
     p->needs_recreate = false;
     p->cur_width = sinfo.imageExtent.width;
     p->cur_height = sinfo.imageExtent.height;
 
-    // Freeing the old swapchain while it's still in use is an error, so do it
-    // asynchronously once the device is idle
-    if (sinfo.oldSwapchain) {
-        p->old_swapchain = sinfo.oldSwapchain;
-        vk_dev_callback(vk, (vk_cb) destroy_swapchain, vk, p);
-    }
+    PL_DEBUG(sw, "(Re)creating swapchain of size %dx%d",
+             sinfo.imageExtent.width,
+             sinfo.imageExtent.height);
+
+    // Calling `vkCreateSwapchainKHR` puts sinfo.oldSwapchain into a retired
+    // state whether the call succeeds or not, so we always need to garbage
+    // collect it afterwards - asynchronously as it may still be in use
+    sinfo.oldSwapchain = p->swapchain;
+    p->swapchain = VK_NULL_HANDLE;
+    VkResult res = vk->CreateSwapchainKHR(vk->dev, &sinfo, PL_VK_ALLOC, &p->swapchain);
+    vk_dev_callback(vk, (vk_cb) destroy_swapchain, vk, vk_wrap_handle(sinfo.oldSwapchain));
+    PL_VK_ASSERT(res, "vk->CreateSwapchainKHR(...)");
 
     // Get the new swapchain images
     VK(vk->GetSwapchainImagesKHR(vk->dev, p->swapchain, &num_images, NULL));
@@ -604,7 +608,7 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
 
     // If needed, allocate some more semaphores
     while (num_images > p->sems.num) {
-        VkSemaphore sem_in = NULL, sem_out = NULL;
+        VkSemaphore sem_in = VK_NULL_HANDLE, sem_out = VK_NULL_HANDLE;
         static const VkSemaphoreCreateInfo seminfo = {
             .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
         };
@@ -652,7 +656,7 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
     p->color_repr.bits.sample_depth = bits;
     p->color_repr.bits.color_depth = bits;
 
-    // FIXME: infer `p->color_space.sig_peak` etc. from HDR metadata?
+    // Note: `p->color_space.hdr` is (re-)applied by `set_hdr_metadata`
     map_color_space(sinfo.imageColorSpace, &p->color_space);
 
     // Forcibly re-apply HDR metadata, bypassing the no-op check
@@ -666,13 +670,9 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
 error:
     PL_ERR(vk, "Failed (re)creating swapchain!");
     pl_free(vkimages);
-    if (p->swapchain != sinfo.oldSwapchain) {
-        vk->DestroySwapchainKHR(vk->dev, p->swapchain, PL_VK_ALLOC);
-        p->swapchain = VK_NULL_HANDLE;
-        p->cur_width = p->cur_height = 0;
-        p->suboptimal = false;
-        p->needs_recreate = false;
-    }
+    vk->DestroySwapchainKHR(vk->dev, p->swapchain, PL_VK_ALLOC);
+    p->swapchain = VK_NULL_HANDLE;
+    p->cur_width = p->cur_height = 0;
     return false;
 }
 
@@ -693,7 +693,7 @@ static bool vk_sw_start_frame(pl_swapchain sw,
     }
 
     VkSemaphore sem_in = p->sems.elem[p->idx_sems].in;
-    PL_TRACE(vk, "vkAcquireNextImageKHR signals %p", (void *) sem_in);
+    PL_TRACE(vk, "vkAcquireNextImageKHR signals 0x%"PRIx64, (uint64_t) sem_in);
 
     for (int attempts = 0; attempts < 2; attempts++) {
         uint32_t imgidx = 0;
@@ -795,7 +795,7 @@ static bool vk_sw_submit_frame(pl_swapchain sw)
         .pImageIndices = &idx,
     };
 
-    PL_TRACE(vk, "vkQueuePresentKHR waits on %p", (void *) sem_out);
+    PL_TRACE(vk, "vkQueuePresentKHR waits on 0x%"PRIx64, (uint64_t) sem_out);
     vk->lock_queue(vk->queue_ctx, pool->qf, qidx);
     VkResult res = vk->QueuePresentKHR(queue, &pinfo);
     vk->unlock_queue(vk->queue_ctx, pool->qf, qidx);
